// CS 106X Fall 2018 - Nick Troccoli
// C++ solutions for problems from Section #3 (Week 4)
// Section topic: Recursive Backtracking

#include <iostream>
#include <fstream>
#include "console.h"
#include "stack.h"
#include "set.h"
#include "grid.h"
#include "vector.h"
#include "lexicon.h"

using namespace std;

/*****************************************************
 * 1. longestCommonSubsequence - (exhaustive search) *
 *****************************************************/
string longestCommonSubsequence(const string& s1, const string& s2) {
    if (s1.length() == 0 || s2.length() == 0) {
        return "";
    } else if (s1[0] == s2[0]) {
        return s1[0] + longestCommonSubsequence(s1.substr(1), s2.substr(1));
    } else {
        string choice1 = longestCommonSubsequence(s1, s2.substr(1));
        string choice2 = longestCommonSubsequence(s1.substr(1), s2);
        if (choice1.length() >= choice2.length()) {
            return choice1;
        } else {
            return choice2;
        }
    }
}


/***************************************
 * 2. floodFill - (depth-first search) *
 ***************************************/
int floodFillHelper(const Grid<double>& elevs, int row, int col, double prevElevation, Grid<bool>& flooded) {
    if (!elevs.inBounds(row, col) || flooded[row][col]) {
        // base case: out of bounds or already counted this square
        return 0;
    } else if (elevs[row][col] > prevElevation) {
        // base case: water can’t flow upward.
        return 0;
    }
    // recursive case: water flows to neighbors; total impact is this square plus impact on
    // neighboring regions
    flooded[row][col] = true;
    return 1 + floodFillHelper(elevs, row - 1, col, elevs[row][col], flooded)  //north
             + floodFillHelper(elevs, row + 1, col, elevs[row][col], flooded)  //south
             + floodFillHelper(elevs, row, col + 1, elevs[row][col], flooded)  //east
             + floodFillHelper(elevs, row, col - 1, elevs[row][col], flooded); //west
}

int floodFill(const Grid<double>& elevs, int row, int col) {
    Grid<bool> flooded(elevs.numRows(), elevs.numCols(), false);
    return floodFillHelper(elevs, row, col, elevs[row][col], flooded);
}


/*************************************
 * 3. partitionable - (backtracking) *
 *************************************/
bool partitionableHelper(Vector<int>& rest, int sum1, int sum2) {
    if (rest.isEmpty()) {
        return sum1 == sum2;
    } else {
        int n = rest[0];
        rest.remove(0);
        bool answer = partitionableHelper(rest, sum1 + n, sum2)
                   || partitionableHelper(rest, sum1, sum2 + n);
        rest.insert(0, n);
        return answer;
    }
}

bool partitionable(Vector<int>& v) {
    return partitionableHelper(v, 0, 0);
}


/**********************************
 * 4. makeChange - (backtracking) *
 **********************************/
void makeChangeHelper(int amount, Vector<int>& denominations, Vector<int>& chosen) {
    if (denominations.isEmpty()) {
        if (amount == 0) {
            cout << chosen << endl;
        }
    } else {
        int denomination = denominations[0];
        denominations.remove(0);
        for (int i = 0; i <= (amount / denomination); i++) {
            chosen += i;
            makeChangeHelper(amount - (i * denomination), denominations, chosen);
            chosen.remove(chosen.size() - 1);
        }
        denominations.insert(0, denomination);
    }
}

void makeChange(int amount, Vector<int>& denominations) {
    Vector<int> chosen;
    makeChangeHelper(amount, denominations, chosen);
}


/************************************
 * 5. printSquares - (backtracking) *
 ************************************/

// Prints well-formatted output for problem 6 using standard C++ iterators,
// for those who are curious
void displaySquaredSum(const Set<int>& numsToBeSquared) {
    auto elem = numsToBeSquared.begin();
    cout << *elem << "^2";
    while (++elem != numsToBeSquared.end()) {
        cout << " + " << *elem << "^2";
    }
    cout << endl;
}

// Explore all ways to form n as a sum of squares of integers starting
// with the given min and storing the chosen results in the given set.
void printSquaresHelper(int n, int min, Set<int>& chosen) {
    if (n == 0) {
        displaySquaredSum(chosen);   // base case: sum has reached n
    } else {
        // recursive case: try all combinations of every integer
        int max = (int) sqrt(n); // valid choices go up to sqrt(n)
        for (int i = min; i <= max; i++) {
            // try all combinations that include the square of this integer
            chosen.add(i);
            printSquaresHelper(n - (i * i), i + 1, chosen);
            chosen.remove(i); // backtrack
        }
    }
}

// Prints all ways to express n as a sum of squares of unique integers.
// Precondition: n >= 0
void printSquares(int n) {
    Set<int> chosen;
    printSquaresHelper(n, 1, chosen);
}


/**********************************
 * 6. listTwiddles (backtracking) *
 **********************************/
void listTwiddlesHelper(const string& prefix, const string& str, int index, const Lexicon& lex) {
    if (!lex.containsPrefix(prefix)) {
        return;   // optimization; not strictly necessary but good to do
    } else if (index >= (int)str.size()) {
        if (lex.contains(prefix)) {
            cout << prefix << endl;
        }
    } else {
        for (char ch = str[index] - 2; ch <= str[index] + 2; ch++) {
            if (isalpha(ch)) {
                listTwiddlesHelper(prefix + ch, str, index + 1, lex);
            }
        }
    }
}

void listTwiddles(const string& str, const Lexicon& lex) {
    string prefix = "";
    listTwiddlesHelper(prefix, str, /* index */ 0, lex);
}


/***********************************
 * 7. waysToClimb - (backtracking) *
 ***********************************/
void waysToClimbHelper(int numStairs, Stack<int>& chosen) {
    if (numStairs <= 0) {
        cout << chosen << endl;
    } else {
        chosen.push(1);                                // choose 1
        waysToClimbHelper(numStairs - 1, chosen);      // explore
        chosen.pop();                                  // un-choose

        if (numStairs > 1) {
            chosen.push(2);                            // choose 2
            waysToClimbHelper(numStairs - 2, chosen);  // explore
            chosen.pop(); // un-choose
        }
    }
}

// Precondition: numStairs >= 0
void waysToClimb(int numStairs) {
    Stack<int> chosen;
    waysToClimbHelper(numStairs, chosen);
}


/****************************************
 * 8. XLSolutionExists - (backtracking) *
 ****************************************/

// See separate cs106x-section3-xlup-puzzle-solution Qt project for problem 9.


/* main function for testing individual problems */

int main() {
    cout << "Problem 1:" << endl;
    cout << longestCommonSubsequence("nick", "zach") << endl;
    cout << longestCommonSubsequence("nick", "106x") << endl;
    cout << longestCommonSubsequence("she sells", "seashells") << endl;
    
    cout << endl << "Problem 2:" << endl;           //Flooded squares marked with X for water starting at (0, 2)
    Grid<double> terrain{{5.0, 4.5, 4.5, 3.0},      // - X X X
                         {2.5, 5.0, 3.5, 3.0},      // - - X X
                         {3.0, 1.5, 4.0, 2.0},      // - - - X
                         {0.5, 1.0, 1.5, 2.0}};     // X X X X
    cout << floodFill(terrain, 0, 2) << endl;       // for a total of 10 flooded squares

    cout << endl << "Problem 3:" << endl;
    Vector<int> v1 {1, 1, 2, 3, 5};
    cout << partitionable(v1) << endl;
    Vector<int> v2{1, 4, 5, 6};
    cout << partitionable(v2) << endl;

    cout << endl << "Problem 4:" << endl;
    Vector<int> exampleCoins{1, 5, 10};
    makeChange(15, exampleCoins);
    
    cout << endl << "Problem 5:" << endl;
    printSquares(200);
    
    cout << endl << "Problem 6:" << endl;
    Lexicon english("EnglishWords.dat");
    cout << "For sparks:" << endl;
    listTwiddles("sparks", english);
    cout << "For craggy:" << endl;
    listTwiddles("craggy", english);
    
    cout << endl << "Problem 7:" << endl;
    waysToClimb(4);
    
    return 0;
}
